UMLS_REST_ENDPOINT = "https://uts-ws.nlm.nih.gov/rest"

API_KEY_ENV_VAR = "UMLS_API_KEY"

# Common vocabulary name mapping
_COMMON_VOCAB_NAMES = {
    "SNOMED": "SNOMEDCT_US",
    "SNOMED-CT": "SNOMEDCT_US",
    "LOINC": "LNC",
}
